"""Run MQIQC."""
import argparse
from os import system, popen
from pathlib import Path
import json

from pandas import read_csv

description_template = dict(
    BIDSVersion="1.0.0",
    License="",
    Name="MRI dataset",
    ReferencesAndLinks=["https://github.com/garciadias/Neuroharmony"],
)


def _run_MRIQC_on_dir(subj_dir, sub_out, n_jobs=5):
    """Run MQIQC."""
    system("docker pull poldracklab/mriqc:0.15.2rc1")
    docker_command = (
        f"docker run -it --rm -v {subj_dir}/:/data:ro -v {sub_out}/:/out "
        "poldracklab/mriqc:0.15.2rc1 /data /out participant --participant_label"
    )
    subject_list = _get_subject_list(subj_dir, sub_out)
    for i in range(len(subject_list) // n_jobs + 1):
        run_str = docker_command
        for subj in subject_list[(i * n_jobs): ((i + 1) * n_jobs)]:
            run_str += " " + subj
        system(run_str + (f" --n_procs {n_jobs} --no-sub"))


def _get_subject_list(subj_dir, sub_out):
    """Get subjects."""
    subj_list = [subj.name.split("-")[1] for subj in subj_dir.glob("sub-*")]
    subj_done = [subj.name.split("-")[1] for subj in sub_out.glob("sub-*") if subj.is_dir()]
    return [subj for subj in subj_list if subj not in subj_done]


def _verify_docker_is_installed():
    if len(popen("command -v docker").read()) == 0:
        raise EnvironmentError("Docker is not installed. Get docker at https://docs.docker.com/get-docker/")


def run_mriqc(subj_dir, out_dir, n_jobs=1):
    """Run MRIQC.

    .. _MRIQC: https://github.com/poldracklab/mriqc
    .. _`MRIQC documentation`: https://github.com/poldracklab/mriqc
    .. _`docker`: https://docs.docker.com/get-docker/
    This function is a wrapper to run the MRIQC_ tool on a docker_ container. It requires docker to be installed. MRIQC_
    extracts no-reference IQMs (image quality metrics) from structural (T1w and T2w) and functional MRI (magnetic
    resonance imaging) data. MRIQC is an open-source project. More information can be found at `MRIQC documentation`_.

    Parameters
    ----------
    subj_dir: str
     Path to BIDS directory.
    out_dir: str
     Path to the output of MRIQC analysis.
    n_jobs: int
     Number of cpu used to run MRIQC.

    Returns
    -------
    IQMs: NDFrame of shape [n_subjects, 68]
     Dataframe with the IQMs for each subject.
    """
    _verify_docker_is_installed()
    subj_dir = Path(subj_dir)
    Path(f"{out_dir}/{subj_dir.name}").mkdir(exist_ok=True, parents=True)
    sub_out = Path(f"{out_dir}/{subj_dir.name}/").absolute()
    if not Path(f"{subj_dir}/dataset_description.json").exists():
        with open(f"{subj_dir}/dataset_description.json", "w") as description_file:
            json.dump(description_template, description_file)
    _run_MRIQC_on_dir(subj_dir.absolute(), sub_out, n_jobs=n_jobs)
    get_group_cmd = (
        f"docker run -it --rm -v {sub_out}/:/data:ro -v {sub_out}/:/out poldracklab/mriqc:0.15.2rc1" " /data /out group"
    )
    get_pred_cmd = (
        f"docker run -v {sub_out}:/scratch -w /scratch --entrypoint=mriqc_clf poldracklab/mriqc:0.15.2rc1"
        " --load-classifier -X group_T1w.tsv"
    )
    system(get_group_cmd)
    system(get_pred_cmd)
    return read_csv(f"{out_dir}/group_T1w.tsv", delimiter="\t", index_col=0)


def init_argparse():
    description = (
        "A Neuroharmony wrapper to run the docker version of the MRI quality control tool by"
        "Esteban et al., (2017). MRIQC: Advancing the automatic prediction of image quality in MRI from unseen sites."
        " https://github.com/poldracklab/mriqc"
    )

    parser = argparse.ArgumentParser(usage="mriqc-run [subj_dir] [out_dir] [OPTIONS]", description=description)

    parser.add_argument("-v", "--version", action="version", version="mriqc-run version 1.0.0")

    parser.add_argument("subj_dir", action="store", type=str, help="Path to BIDS directory.")
    parser.add_argument("out_dir", action="store", type=str, help="Path to the output of MRIQC analysis.")
    parser.add_argument(
        "-n_jobs",
        action="store",
        default=1,
        type=int,
        help="Number of cpu used in processing the images. Default is 1.",
    )

    return parser


def main():
    parser = init_argparse()
    args = parser.parse_args()
    if popen("command -v docker").read():
        run_mriqc(args.subj_dir, args.out_dir, args.n_jobs)
    else:
        raise OSError("Docker is not installed. Get docker at https://docs.docker.com/get-docker/")



"""Test tools for data finding and collection."""
from collections import namedtuple
from pathlib import Path
from os import popen

import pytest

from neuroharmony.data import collect_tools
from pandas.core.generic import NDFrame


def test_fetch_sample():
    data = collect_tools.fetch_sample()
    assert isinstance(data, NDFrame)



"""Tests for harmonization with Neuroharmony."""
from collections import namedtuple

from pandas import concat
from pandas.core.generic import NDFrame
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
import joblib
import pytest

from neuroharmony import fetch_trained_model
from neuroharmony.data.collect_tools import fetch_sample
from neuroharmony.data.rois import rois
from neuroharmony.models.harmonization import exclude_single_subject_groups
from neuroharmony.models.harmonization import Neuroharmony, _label_encode_covariates, _label_decode_covariates
from neuroharmony.models.metrics import ks_test_grid


@pytest.fixture(scope="session")
def resources(tmpdir_factory):
    """Set up."""
    r = namedtuple("resources", "data_path")
    r.features = rois[:3]
    r.regression_features = [
        "Age",
        "summary_gm_median",
        "spacing_x",
        "summary_gm_p95",
        "cnr",
        "size_x",
        "cjv",
        "summary_wm_mean",
        "icvs_gm",
        "wm2max",
    ]
    r.covariates = ["Gender", "scanner", "Age"]
    r.eliminate_variance = ["scanner"]
    r.original_data = fetch_sample()
    exclude_vars = r.original_data.columns[r.original_data.isna().sum() != 0].to_list()
    r.original_data = r.original_data[[var for var in r.original_data.columns if var not in exclude_vars]]
    r.original_data = r.original_data[~r.original_data[r.covariates].isna().any(axis=1)]
    r.original_data.Age = r.original_data.Age.astype(int)
    scanners = r.original_data.scanner.unique()
    train_bool = r.original_data.scanner.isin(scanners[1:])
    test_bool = r.original_data.scanner.isin(scanners[:1])
    r.X_train_split = r.original_data[train_bool]
    r.X_train_split = exclude_single_subject_groups(r.X_train_split, r.covariates)
    r.X_test_split = r.original_data[test_bool]
    r.n_scanners = len(r.original_data.scanner.unique())
    return r


@pytest.fixture(scope="session")
def model(resources):
    """Define mock model.

    Parameters
    ----------
    resources : namedtuple
        Basic information needed to build a neuroharmony model.

    Returns
    -------
    Neuroharmony class
        Neuroharmony model.
    """
    neuroharmony = Neuroharmony(
        resources.features,
        resources.regression_features,
        resources.covariates,
        resources.eliminate_variance,
        param_distributions=dict(
            RandomForestRegressor__n_estimators=[5, 10, 15, 20],
            RandomForestRegressor__random_state=[42, 78],
            RandomForestRegressor__warm_start=[False, True],
        ),
        estimator_args=dict(n_jobs=1, random_state=42),
        randomized_search_args=dict(cv=5, n_jobs=27),
    )
    return neuroharmony


def test_label_encode_decode(resources):
    """Test encoder and decoder."""
    df, encoders = _label_encode_covariates(resources.X_train_split, resources.covariates)
    assert all([isinstance(value, int) for value in df.scanner])
    df = _label_decode_covariates(df, resources.covariates, encoders)
    assert all([isinstance(value, str) for value in df.scanner])


@pytest.fixture
def test_neuroharmony_behaviour(resources):
    """Test Neuroharmony."""
    x_train, x_test = resources.X_train_split, resources.X_test_split
    neuroharmony = Neuroharmony(
        resources.features,
        resources.regression_features,
        resources.covariates,
        resources.eliminate_variance,
        param_distributions=dict(
            RandomForestRegressor__n_estimators=[5, 10, 15, 20],
            RandomForestRegressor__random_state=[42, 78],
            RandomForestRegressor__warm_start=[False, True],
        ),
        estimator_args=dict(n_jobs=1, random_state=42),
        randomized_search_args=dict(cv=5, n_jobs=27),
    )
    x_train_harmonized = neuroharmony.fit_transform(x_train)
    x_test_harmonized = neuroharmony.predict(x_test)
    joblib.dump(neuroharmony, "data/neuroharmony.pkl.gz")
    data_harmonized = concat([x_train_harmonized, x_test_harmonized], sort=False)
    KS_original = ks_test_grid(resources.original_data, resources.features, "scanner")
    KS_harmonized = ks_test_grid(data_harmonized, resources.features, "scanner")
    assert KS_original[resources.features[0]].shape == (resources.n_scanners, resources.n_scanners)
    assert KS_harmonized[resources.features[0]].shape == (resources.n_scanners, resources.n_scanners)
    assert isinstance(x_test, NDFrame)
    assert isinstance(neuroharmony, BaseEstimator)
    assert not neuroharmony.prediction_is_covered_.all(), "No subjects out of the range."


def test_predict_untrained_model(model, resources):
    """Test check model can record the training range of each variables."""
    neuroharmony = model
    _, x_test = resources.X_train_split, resources.X_test_split
    with pytest.raises(NotFittedError):
        neuroharmony.predict(x_test)


def test_ckeck_training_range(model, resources):
    """Test check model can record the training range of each variables."""
    neuroharmony = model
    neuroharmony._check_training_ranges(resources.X_train_split)
    assert isinstance(neuroharmony.coverage_, NDFrame), "coverage_ is not DataFrame."
    assert not neuroharmony.coverage_.isna().any().any(), "NaN field detected."


def test_ckeck_prediction_range(model, resources):
    """Test we can verify if the prediction sample is covered by the training ranges."""
    neuroharmony = model
    neuroharmony._check_training_ranges(resources.X_train_split)
    neuroharmony._check_prediction_ranges(resources.X_test_split)
    assert isinstance(neuroharmony.prediction_is_covered_, NDFrame), "prediction_is_covered_ is not DataFrame."
    assert not neuroharmony.prediction_is_covered_.isna().any(), "NaN field detected."
    assert not neuroharmony.prediction_is_covered_.all(), "No subjects out of the range."
    assert isinstance(neuroharmony.subjects_out_of_range_, list), "The subjects_out_of_range_ is not a list."


def test_fetch_model(test_neuroharmony_behaviour):
    """Test a trained model can be retrained."""
    neuroharmony = fetch_trained_model()
    assert isinstance(neuroharmony.coverage_, NDFrame)


def test_retrain_a_model(test_neuroharmony_behaviour, resources):
    """Test a trained model can be retrained."""
    neuroharmony = joblib.load("data/neuroharmony.pkl.gz")
    x_train, _ = resources.X_train_split, resources.X_test_split
    neuroharmony.refit(x_train)
    pass



"""Test the combination of the raw data files."""
from collections import namedtuple

from pandas.core.generic import NDFrame
from pathlib import Path, PosixPath
import pytest

from neuroharmony.data import combine_tools, collect_tools


@pytest.fixture(scope="session")
def resources(tmpdir_factory):
    """Set up."""
    r = namedtuple("resources", "data_root")
    r.data_root = "./data"
    r.raw_root = "%s/raw/" % r.data_root
    r.dataset_list = list(collect_tools.find_all_files_by_name("data/raw/", "*", depth=1))
    return r


def test_get_scanners(resources):
    """Test we can get the scanners from a single site."""
    site = combine_tools.Site(resources.dataset_list[0])
    assert isinstance(site.scanner_list, list)


def test_get_files(resources):
    """Test we can get the files for each scanner."""
    site = combine_tools.Site(Path("data/raw/COBRE"))
    assert isinstance(site.SCANNER01.freesurferData_path, PosixPath)
    site = combine_tools.Site(Path("data/raw/IXI"))
    assert isinstance(site.SCANNER01.freesurferData_path, PosixPath)
    assert isinstance(site.SCANNER01.participants_path, PosixPath)
    assert isinstance(site.SCANNER01.iqm_path, PosixPath)
    assert isinstance(site.SCANNER02.pred_path, PosixPath)
    assert isinstance(site.SCANNER03.qoala_path, PosixPath)


def test_load_files():
    """Test the files exist and can be open as NDFrames."""
    site = combine_tools.Site(Path("data/raw/COBRE"))
    assert isinstance(site.SCANNER01.freesurferData, NDFrame)
    assert isinstance(site.SCANNER01.participants, NDFrame)
    assert isinstance(site.SCANNER01.iqm, NDFrame)
    assert isinstance(site.SCANNER01.pred, NDFrame)
    assert isinstance(site.SCANNER01.qoala, NDFrame)


def test_combine_freesurfer():
    """Test the combination of the freesurfer output return a dataframe."""
    mri_path = collect_tools.fetch_mri_data()
    FREESURFER = combine_tools.combine_freesurfer(f"{mri_path}/derivatives/freesurfer/")
    assert isinstance(FREESURFER, NDFrame)
    assert FREESURFER.index[0] == "sub-0001"


def test_combine_mriqc():
    mri_path = collect_tools.fetch_mri_data()
    print(mri_path)
    MRIQC = combine_tools.combine_mriqc(f"{mri_path}/derivatives/mriqc/")
    assert isinstance(MRIQC, NDFrame)
    assert MRIQC.index[0] == "sub-0001"



"""Collect tools."""
from pathlib import Path
from requests import get
from shutil import copyfile
from zipfile import ZipFile
import joblib
import os

from pandas import read_csv
from tqdm import tqdm


def _download(url, filepath):
    dirpath = Path(filepath).parent
    Path(dirpath).mkdir(exist_ok=True)
    headers = {"user-agent": "Wget/1.16 (linux-gnu)"}
    r = get(url, stream=True, headers=headers)
    total_size = int(r.headers.get("content-length", 0))
    block_size = 1024
    t = tqdm(total=total_size, unit="iB", unit_scale=True)
    with open(filepath, "wb") as f:
        for data in r.iter_content(block_size):
            t.update(len(data))
            f.write(data)
    t.close()


def fetch_mri_data():
    """Fetch example of MRI dataset.

    The dataset is a replication of the Bert subject released with the FreeSurfer software for testing.

    Returns
    =======
    mri_path: str
        The path for the MRI data.
    """
    script_path = os.path.dirname(os.path.abspath(__file__))
    filepath = f"{script_path}/../../data/mri.zip"
    unzip_folder = str(Path(filepath).parent) + "/mri/"
    if not Path(filepath).exists():
        _download("https://www.dropbox.com/s/kcbq0266bcab3bx/ds002936.zip", filepath)
        Path(unzip_folder).mkdir(exist_ok=True)
        zip_file = ZipFile(filepath, "r")
        zip_file.extractall(unzip_folder)
        zip_file.close()
        os.remove(filepath)
    return str(Path(unzip_folder).absolute())


def fetch_sample():
    """Fetch a sample of FreeSurfer derived volumes in the Neuroharmony format.

    Fetch the FreeSurfer derived volumes of some subjects in the
    `ADHD200 <http://fcon_1000.projects.nitrc.org/indi/adhd200/index.html>`_ and in
    the `PPMI <http://www.ppmi-info.org/>`_ datasets.

    Returns
    =======
    dataset: NDFrame of shape [n_subjects, n_features]
        DataFrame with data from ADHD200 and the PPMI subjects in the  Neuroharmony format.
    """
    script_path = os.path.dirname(os.path.abspath(__file__))
    filepath = f"{script_path}/../../data/test_sample.csv"
    _download("https://www.dropbox.com/s/mxcaqx2y29n09rp/test_sample.csv", filepath)
    return read_csv(filepath, index_col=0)


def fetch_trained_model():
    """Fetch Neuroharmony pre-trained model.

    Returns
    =======
    neuroharmony: Neuroharmony class
        Pre-trained Neuroharmony model.
    """
    script_path = os.path.dirname(os.path.abspath(__file__))
    filepath = f"{script_path}/../../data/neuroharmony.pkl.gz"
    if not Path(filepath).exists():
        _download("https://www.dropbox.com/s/s3521oqd3fpi0ll/neuroharmony.pkl.gz", filepath)
    try:
        return joblib.load(filepath)
    except (KeyError, ValueError):
        Path(filepath).unlinke()
        _download("https://www.dropbox.com/s/s3521oqd3fpi0ll/neuroharmony.pkl.gz", filepath)
        return joblib.load(filepath)


def find_all_files_by_name(directory_path, file_pattern, depth=2):
    """Find all files in a folder.

    Parameters
    ==========
    directory_path: string
        The path for a giving folder.

    file_pattern: string
        File extension (.csv, .pdf, .txt, ...).

    depth: int
        Depth of the file search.

    Returns
    =======
    filelist: list
        List of paths to the found files in the format of pathlib.PosixPath.
    """
    filelist = []
    for level in range(depth):
        filelist.extend(list(Path(directory_path).glob("/".join(level * ["*"] + [file_pattern]))))
    return filelist


def collect_datafile(filepath, root_path, local_path):
    """Collect a datafile.

    Parameters
    ==========
    filepath: string or pathlib.PosixPath
        The path for the file to be copied.

    root_path: string
        The path root to the origin of the data.

    local_path: string
        The path to the local folder you want to save the copied data.

    Returns
    =======
    file_exists: boolean
        Returns True if the files were copied correctly and False otherwise.
    """
    filepath = str(filepath)
    local_final_path = filepath.replace(root_path, local_path)
    Path(local_final_path).parent.mkdir(parents=True, exist_ok=True)
    copyfile(filepath, local_final_path)
    return Path(local_final_path).exists()


def collect_multiple_datafile(filepath_list, root_path, local_path):
    """Collect a list of datafiles.

    Parameters
    ==========
    filepath_list: list of strings or pathlib.PosixPath
        List of paths for the file to be copied.

    root_path: string
        The path root to the origin of the data.

    local_path: string
        The path to the local folder you want to save the copied data.

    Returns
    =======
    file_exists: boolean
        Returns True if the files were copied correctly and False otherwise.
    """
    for filepath in filepath_list:
        collect_datafile(filepath, root_path, local_path)


if __name__ == "__main__":
    SERVER_ROOT = "/media/kcl_2/HDD/SynologyDrive"
    PARTICIPANT_ROOT = "%s/BIDS_data/" % SERVER_ROOT
    FREESURFER_ROOT = "%s/FreeSurfer_preprocessed/" % SERVER_ROOT
    QOALA_ROOT = "%s/Qoala/" % SERVER_ROOT
    MRIQC_ROOT = "%s/MRIQC/" % SERVER_ROOT
    Path("./data/processed").mkdir(exist_ok=True, parents=True)
    PARTICIPANTS_FILES = find_all_files_by_name(PARTICIPANT_ROOT, "participants.tsv", depth=3)
    for file_path in PARTICIPANTS_FILES:
        df = read_csv(file_path, header=0, sep="\t")
        df["image_id"] = df[["participant_id", "session_id", "acq_id", "run_id"]].agg("_".join, axis=1) + "_T1w"
        df.to_csv(file_path, index=False, sep="\t")
    FSURFER_FILES = find_all_files_by_name(FREESURFER_ROOT, "freesurferData.csv", depth=3)
    QOALA_FILES = find_all_files_by_name(QOALA_ROOT, "Qoala*.csv", depth=3)
    MRIQC_GROUP_FILES = find_all_files_by_name(MRIQC_ROOT, "group_T1w.tsv", depth=3)
    MRIQC_PRED_FILES = find_all_files_by_name(MRIQC_ROOT, "*pred.csv", depth=3)
    collect_multiple_datafile(PARTICIPANTS_FILES, PARTICIPANT_ROOT, "./data/raw/")
    collect_multiple_datafile(FSURFER_FILES, FREESURFER_ROOT, "./data/raw/")
    collect_multiple_datafile(QOALA_FILES, QOALA_ROOT, "./data/raw/")
    collect_multiple_datafile(MRIQC_GROUP_FILES, MRIQC_ROOT, "./data/raw/")
    collect_multiple_datafile(MRIQC_PRED_FILES, MRIQC_ROOT, "./data/raw/")



"""List of Regions of interest."""
rois = ['Left-Lateral-Ventricle',
        'Left-Inf-Lat-Vent',
        'Left-Cerebellum-White-Matter',
        'Left-Cerebellum-Cortex',
        'Left-Thalamus-Proper',
        'Left-Caudate',
        'Left-Putamen',
        'Left-Pallidum',
        '3rd-Ventricle',
        '4th-Ventricle',
        'Brain-Stem',
        'Left-Hippocampus',
        'Left-Amygdala',
        'CSF',
        'Left-Accumbens-area',
        'Left-VentralDC',
        'Right-Lateral-Ventricle',
        'Right-Inf-Lat-Vent',
        'Right-Cerebellum-White-Matter',
        'Right-Cerebellum-Cortex',
        'Right-Thalamus-Proper',
        'Right-Caudate',
        'Right-Putamen',
        'Right-Pallidum',
        'Right-Hippocampus',
        'Right-Amygdala',
        'Right-Accumbens-area',
        'Right-VentralDC',
        'CC_Posterior',
        'CC_Mid_Posterior',
        'CC_Central',
        'CC_Mid_Anterior',
        'CC_Anterior',
        'lh_bankssts_volume',
        'lh_caudalanteriorcingulate_volume',
        'lh_caudalmiddlefrontal_volume',
        'lh_cuneus_volume',
        'lh_entorhinal_volume',
        'lh_fusiform_volume',
        'lh_inferiorparietal_volume',
        'lh_inferiortemporal_volume',
        'lh_isthmuscingulate_volume',
        'lh_lateraloccipital_volume',
        'lh_lateralorbitofrontal_volume',
        'lh_lingual_volume',
        'lh_medialorbitofrontal_volume',
        'lh_middletemporal_volume',
        'lh_parahippocampal_volume',
        'lh_paracentral_volume',
        'lh_parsopercularis_volume',
        'lh_parsorbitalis_volume',
        'lh_parstriangularis_volume',
        'lh_pericalcarine_volume',
        'lh_postcentral_volume',
        'lh_posteriorcingulate_volume',
        'lh_precentral_volume',
        'lh_precuneus_volume',
        'lh_rostralanteriorcingulate_volume',
        'lh_rostralmiddlefrontal_volume',
        'lh_superiorfrontal_volume',
        'lh_superiorparietal_volume',
        'lh_superiortemporal_volume',
        'lh_supramarginal_volume',
        'lh_frontalpole_volume',
        'lh_temporalpole_volume',
        'lh_transversetemporal_volume',
        'lh_insula_volume',
        'rh_bankssts_volume',
        'rh_caudalanteriorcingulate_volume',
        'rh_caudalmiddlefrontal_volume',
        'rh_cuneus_volume',
        'rh_entorhinal_volume',
        'rh_fusiform_volume',
        'rh_inferiorparietal_volume',
        'rh_inferiortemporal_volume',
        'rh_isthmuscingulate_volume',
        'rh_lateraloccipital_volume',
        'rh_lateralorbitofrontal_volume',
        'rh_lingual_volume',
        'rh_medialorbitofrontal_volume',
        'rh_middletemporal_volume',
        'rh_parahippocampal_volume',
        'rh_paracentral_volume',
        'rh_parsopercularis_volume',
        'rh_parsorbitalis_volume',
        'rh_parstriangularis_volume',
        'rh_pericalcarine_volume',
        'rh_postcentral_volume',
        'rh_posteriorcingulate_volume',
        'rh_precentral_volume',
        'rh_precuneus_volume',
        'rh_rostralanteriorcingulate_volume',
        'rh_rostralmiddlefrontal_volume',
        'rh_superiorfrontal_volume',
        'rh_superiorparietal_volume',
        'rh_superiortemporal_volume',
        'rh_supramarginal_volume',
        'rh_frontalpole_volume',
        'rh_temporalpole_volume',
        'rh_transversetemporal_volume',
        'rh_insula_volume', ]



"""Metrics for evaluation of the normalization."""
from __future__ import (nested_scopes, generators, division, absolute_import,
                        with_statement, print_function, unicode_literals)

from itertools import combinations

import numpy as np
from pandas import DataFrame, Series
from scipy.stats import ks_2samp


def check_vars(df, features):
    """Check if all variables in a list are present in a dataframe."""
    features = Series(features)
    is_feature_present = features.isin(df.columns)
    missing_features_str = "Missing features: %s" % ', '.join(features[~is_feature_present])
    assert is_feature_present.all(), ValueError(missing_features_str)


def ks_test_grid(df, features, sampling_variable='scanner'):
    """Calculate the Kolmogorov-Smirnov score for all pairs of scanners.

    Parameters
    ----------
    df: NDFrame of shape [n_subjects, n_features]
        DataFrame with the subjects data.

    features: list
        List of the features to be considered on the Kolmogorov-Smirnov test.

    sampling_variable: str, default='scanner'
        Variable for which you want to group subjects.

    Returns
    -------
    KS_by_variable: dict of NDFrames
        Kolmogorov-Smirnov p-values to all pairs of instances in the sampling_variable column.
        The keys in the dictionary are the variables in 'features'. The values of each entry are square NDFrames of
        shape [n_vars, n_vars].

    Raises
    ------
    ValueError:
        If the list of variables contain any variable that is not present in df.

    Examples
    --------
    >>> ixi = DataSet('data/raw/IXI').data
    >>> features = ['Left-Lateral-Ventricle', 'Left-Inf-Lat-Vent', ]
    >>> KS = ks_test_grid(df, features, 'scanner')
    >>> KS[features[0]]
    +--------------------------+----------------------+------------------------+--------------------+
    |                          | SCANNER01-SCANNER01  | SCANNER02-SCANNER01    | SCANNER03-SCANNER01|
    +==========================++=====================+========================+====================+
    |SCANNER01-SCANNER01       | NaN                  | NaN                    | NaN                |
    +--------------------------+----------------------+------------------------+--------------------+
    |SCANNER02-SCANNER01       | 0.000759473          | NaN                    | NaN                |
    +--------------------------+----------------------+------------------------+--------------------+
    |SCANNER03-SCANNER01       | 0.0539998            | 0.625887               | NaN                |
    +--------------------------+----------------------+------------------------+--------------------+
    """
    check_vars(df, features)
    groups = df.groupby(sampling_variable)
    scanners_list = df[sampling_variable].unique()
    scanners_list = scanners_list[np.argsort(df[sampling_variable].str.lower().unique())]
    KS_by_variable = {}
    for var in features:
        KS = DataFrame([], index=scanners_list, columns=scanners_list)
        for scanner_batch in np.array_split(np.array(list(combinations(scanners_list, 2))), 80):
            for scanner_a, scanner_b in scanner_batch:
                group_a = groups.get_group(scanner_a)[var]
                group_b = groups.get_group(scanner_b)[var]
                KS.loc[scanner_b][scanner_a] = ks_2samp(group_a, group_b).pvalue
        KS_by_variable[var] = KS
    return KS_by_variable



"""Neuroharmony classes and functions."""
from os import devnull
from warnings import warn
import sys

from tqdm import tqdm
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestRegressor
from sklearn.exceptions import NotFittedError
from sklearn.model_selection import LeaveOneGroupOut, RandomizedSearchCV
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import LabelEncoder, StandardScaler
from pandas.core.generic import NDFrame
from pandas import Series, DataFrame, concat, merge
from numpy import unique, number

from neuroharmony.models.neuroCombat import neuroCombat


def exclude_single_subject_groups(df, covariates):
    """Exclude categories with only one subject.

    Parameters
    ==========
    df : NDFrame of shape [n_subjects, n_features]
        Pandas DataFrame in the Neuroharmony input format containing the variables in the `covariates` list.
    covariates : list
        List of covariates for which the Harmonization should eliminate or conserve.

    Returns
    =======
    df : NDFrame of shape [n_subjects, n_features]
        Padas DataFrame excluding the subjects that would result in a single subject split of the covariates grouping.
    """
    for covar in covariates:
        instances, n = unique(df[covar], return_counts=True)
        category_counts = DataFrame(n, columns=["N"], index=instances)
        single_subj = category_counts[category_counts.N == 1].index.tolist()
        df = concat([df.groupby(covar).get_group(group) for group in df[covar].unique() if group not in single_subj])
    return df


def _supress_print(func):
    """Define decorator for print suppression."""

    def wrapper(*args, **kwargs):
        sys.stdout = open(devnull, "w")
        resuts = func(*args, **kwargs)
        sys.stdout = sys.__stdout__
        return resuts

    return wrapper


def _label_encode_covariates(df, covariates):
    """Encode dataframe covariates."""
    encoders = {}
    for covar in covariates:
        encoders[covar] = LabelEncoder()
        df[covar] = encoders[covar].fit_transform(df[covar])
    return df, encoders


def _label_decode_covariates(df, covariates, encoders):
    """Decode dataframe covariates."""
    for covar in covariates:
        df[covar] = encoders[covar].inverse_transform(df[covar])
    return df


@_supress_print
def combat(*args, **kwargs):
    """Redefine ComBat to suppress printing unhelpful data."""
    return neuroCombat(*args, **kwargs)


class ComBat(BaseEstimator, TransformerMixin):
    """ComBat model for harmonization.

    A wrapper to the `NeuroCombat <https://github.com/ncullen93/neuroCombat>`_ harmonization as described by
    `Fortin et al., (2018) <https://doi.org/10.1016/j.neuroimage.2017.11.024>`_.

    Parameters
    ==========
    features : list
        List of features to be harmonized with ComBat.
    covariates :
        List of covariates for which the variance needs to be conserved.
    eliminate_variance :
        List of variables for which the variance needs to be eliminated.

    Attributes
    ==========
    harmonized_ : NDFrame of shape [n_subjects, n_features]
        ComBat harmonized DataFrame.
    """

    def __init__(self, features, covariates, eliminate_variance):
        """Init class."""
        self.covariates = covariates
        self.eliminate_variance = eliminate_variance
        self.features = features
        self.reindexed = False

    def transform(self, df, y=None):
        """Run ComBat normalization.

        Applies trained model.

        Parameters
        ----------
        df: NDFrame of shape [n_subjects, n_features]
            Dataframe with data for each subject. The dataframe has to contain the features to be harmonized pandas
            the covariates which you want to use to harmonization.

        Returns
        -------
        harmonized: NDFrame of shape [n_subjects, n_features]
            Dataframe with the harmonized data.

        Raises
        ------
        ValueError:
            If there are missing features among the covariates or the features.

        Examples
        --------
        >>> ixi = DataSet('data/raw/IXI').data
        >>> features = ['Left-Lateral-Ventricle', 'Left-Inf-Lat-Vent', ]
        >>> covariates = ['Gender', 'Dataset', 'Age']
        >>> combat = ComBat(features, covariates)
        >>> harmonized = combat(ixi)
        >>> harmonized

        +------------+----------------------+-----------------+------+-------+---+
        |subject_id  |Left-Lateral-Ventricle|Left-Inf-Lat-Vent|Gender|Dataset|Age|
        +============+======================+=================+======+=======+===+
        |sub-002-00-2|  0.007666            |    0.000134     |  0   |   2   |31 |
        +------------+----------------------+-----------------+------+-------+---+
        |sub-012-00-1|  0.003778            |    0.000064     |  1   |   1   |14 |
        +------------+----------------------+-----------------+------+-------+---+
        |sub-013-00-1|  0.012818            |    0.000464     |  1   |   1   |17 |
        +------------+----------------------+-----------------+------+-------+---+
        |sub-014-00-1|  0.006289            |    0.000167     |  0   |   1   |19 |
        +------------+----------------------+-----------------+------+-------+---+
        |sub-015-00-1|  0.003310            |    0.000154     |  1   |   1   |26 |
        +------------+----------------------+-----------------+------+-------+---+
        """
        self._check_data(df.copy())
        self._check_single_subject_groups(df.copy())
        self._check_subjects_with_nans(df.copy())
        df = self._check_index(df.copy())
        df, self.encoders = _label_encode_covariates(df.copy(), unique(self.covariates + self.eliminate_variance))
        self.harmonized_ = self._run_combat(df.copy())
        self.harmonized_ = _label_decode_covariates(
            self.harmonized_, unique(self.covariates + self.eliminate_variance), self.encoders,
        )
        if self.reindexed:
            self.harmonized_ = self._clean_index(self.harmonized_)
        return self.harmonized_

    def fit(self, df):
        """Fit the model.

        Fit all the transforms one after the other and transform the
        data, then fit the transformed data using the final estimator.

        Parameters
        ----------
        df: NDFrame of shape [n_subjects, n_features]
            Training data. Must fulfil input requirements of first step of the pipeline.

        Returns
        -------
        self : ComBat
            This estimator
        """
        self.harmonized_ = self.transform(df.copy())
        return self

    def fit_transform(self, df):
        """Fit to data, then transform it.

        Fits transformer to df and y with optional parameters fit_params and returns a transformed version of df.

        Parameters
        ----------
        df: NDFrame of shape [n_subjects, n_features]
            Training set.

        Returns
        -------
        harmonized_: NDFrame of shape [n_samples, n_features_new]
            Transformed array.
        """
        self.harmonized_ = self.transform(df.copy())
        return self.harmonized_

    def _check_data(self, df):
        type_error = "Input data should be a pandas dataframe (NDFrame)."
        assert isinstance(df, NDFrame), TypeError(type_error)
        self._check_vars(df, self.features)
        self._check_vars(df, self.covariates)
        self._check_vars(df, self.eliminate_variance)

    def _check_index(self, df):
        if df.index.duplicated().sum():
            df.index = [f"{index}_TMP_{i}" for i, index in enumerate(df.index)]
            self.reindexed = True
        return df

    def _clean_index(self, df):
        df.index = [index.split("_TMP_")[0] for index in df.index]
        return df

    def _check_vars(self, df, vars):
        vars = Series(vars)
        is_feature_present = vars.isin(df.columns)
        missing_features_str = "Missing features: %s" % ", ".join(vars[~is_feature_present])
        assert is_feature_present.all(), ValueError(missing_features_str)

    def _reconstruct_original_fieds(self, df, harmonized, extra_vars):
        """Concatenate ComBat data with the original data fields."""
        if df.index.name is not None:
            index_name = df.index.name
        else:
            index_name = "subject_index"
        if df.index.name in df.columns:
            index_name = df.index.name + "_y"
        harmonized = DataFrame(harmonized, index=df.index, columns=self.features)
        harmonized.index.rename(index_name, inplace=True)
        df = df.loc[harmonized.index][extra_vars].reset_index().copy()
        harmonized = harmonized.reset_index()
        return merge(harmonized, df, how="inner", on=index_name).set_index(index_name)

    def _check_single_subject_groups(self, df):
        """Exclude subjects with only a value in the variable field."""
        for covar in self.covariates:
            instances, n = unique(df[covar], return_counts=True)
            category_counts = DataFrame(n, columns=["N"], index=instances)
            single_subj = category_counts[category_counts.N == 1].index.astype("str").tolist()
            if len(single_subj) > 0:
                raise ValueError(
                    "ComBat harmonization requires more than one subject in each split group."
                    f"The following covar imply groups of a single subject: {covar}"
                )

    def _check_subjects_with_nans(self, df):
        if df.isna().any(axis=1).sum() > 0:
            raise ValueError("NaN values found on subjects data.")

    def _run_combat(self, df):
        """Run ComBat for all covariates."""
        extra_vars = df.columns[~df.columns.isin(self.features)]
        harmonized = df.copy()
        for batch_col in self.eliminate_variance:
            harmonized = combat(
                data=harmonized[self.features].copy(), covars=harmonized[self.covariates].copy(), batch_col=batch_col,
            )
            harmonized = self._reconstruct_original_fieds(df, harmonized, extra_vars)
        return harmonized


class Neuroharmony(TransformerMixin, BaseEstimator):
    """Harmonization tool to mitigate scanner bias.

    Parameters
    ----------
    features : list
        Target features to be harmonized, for example, ROIs.
    regression_features : list
        Features used to derive harmonization rules, for example, IQMs.
    covariates : list
        Variables for which we want to eliminate the bias, for example, age, sex, and scanner.
    estimator : sklearn estimator, default=RandomForestRegressor()
        Model to make the harmonization regression.
    scaler : sklearn scaler, default=StandardScaler()
        Scaler used as the first step of the harmonization regression.
    param_distributions : dict, default=dict(RandomForestRegressor__n_estimators=[100, 200, 500],
                                             RandomForestRegressor__warm_start=[False, True], )
        Distribution of parameters to be testes on the RandomizedSearchCV.
    **estimator_args : dict
        Parameters for the estimator.
    **scaler_args : dict
        Parameters for the scaler.
    **randomized_search_args : dict
        Parameters for the RandomizedSearchCV.
        See https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html
    **pipeline_args : dict
        Parameters for the sklearn Pipeline.
        See https://scikit-learn.org/stable/modules/generated/sklearn.pipeline.Pipeline.html

    Attributes
    ----------
    X_harmonized_ : NDFrame [n_subjects, n_features]
        Input data harmonized.
    leaveonegroupout_ :
        Leave One Group Out cross-validator.
    models_by_feature_ :
        Estimators by features.
    """

    def __init__(
        self,
        features,
        regression_features,
        covariates,
        eliminate_variance,
        estimator=RandomForestRegressor(),
        scaler=StandardScaler(),
        decomposition=PCA(),
        param_distributions=dict(
            RandomForestRegressor__n_estimators=[100, 200, 500],
            RandomForestRegressor__criterion=["mse", "mae"],
            RandomForestRegressor__warm_start=[False, True],
        ),
        estimator_args=dict(n_jobs=1, random_state=42, criterion="mae", verbose=False),
        scaler_args=dict(),
        randomized_search_args=dict(),
        pipeline_args=dict(),
    ):
        """Init class."""
        self.covariates = covariates
        self.decomposition = decomposition
        self.eliminate_variance = eliminate_variance
        self.estimator = estimator
        self.estimator.set_params(**estimator_args)
        self.features = features
        self.param_distributions = param_distributions
        self.pipeline_args = pipeline_args
        self.randomized_search_args = randomized_search_args
        self.regression_features = regression_features
        self.reindexed = False
        self.scaler = scaler
        self.scaler.set_params(**scaler_args)

    def fit(self, df):
        """Fit the model.

        Fit all the transforms one after the other and transform the
        data, then fit the transformed data using the final estimator.

        Parameters
        ----------
        df : NDFrame of shape [n_subjects, n_features]
             Training data. Must fulfil input requirements of the first step of the pipeline.

        Returns
        -------
        self : Neuroharmony
               This estimator
        """
        self._check_data(df.copy())
        self._check_training_ranges(df.copy())
        df = self._check_index(df.copy())
        df, self.encoders = _label_encode_covariates(df.copy(), unique(self.covariates + self.eliminate_variance))
        X_train_split, y_train_split = self._run_combat(df.copy())
        self.models_by_feature_ = {}
        desc = "Randomized search of Neuroharmony hyperparameters: "
        for var in tqdm(self.features, desc=desc):
            self.models_by_feature_[var] = self._random_search_with_leave_one_group_out_cv(
                X_train_split[self.regression_features + [var]], y_train_split[var], y_train_split["scanner"],
            )
        return self

    def fit_transform(self, df):
        """Fit to data, then transform it.

        Fits transformer to df and y with optional parameters fit_params
        and returns a transformed version of df.

        Parameters
        ----------
        df: NDFrame of shape [n_subjects, n_features]
            Training set.

        Returns
        -------
        harmonized_: NDFrame of shape [n_samples, n_features_new]
            Data harmonized with ComBat.
        """
        self.fit(df.copy())
        return self.X_harmonized_

    def predict(self, df):
        """Predict regression target for df.

        The predicted regression target of an input sample is computed as the
        mean predicted regression targets of the trees in the forest.

        Parameters
        ----------
        df : NDFrame of shape [n_samples, n_features]
            Pandas dataframe with features, regression_features and covariates.

        Returns
        -------
        y : NDFrame of shape [n_samples, n_features]
            Data harmonized with Neuroharmony.
        """
        # Check data
        self._check_trained_model()
        self._check_data(df.copy())
        self._check_prediction_ranges(df.copy())
        df = self._check_index(df.copy())
        df, self.encoders = _label_encode_covariates(df.copy(), unique(self.covariates + self.eliminate_variance))
        self.models_by_feature_[self.features[0]]._check_is_fitted("predict")
        self.predicted_ = DataFrame([], columns=self.features, index=df.index)
        for var in self.features:
            predicted_y_1 = self.models_by_feature_[var].predict(df[self.regression_features + [var]])
            self.predicted_[var] = df[var] - predicted_y_1
        self.predicted_ = self._reconstruct_original_fieds(df, self.predicted_, self.extra_vars)
        self.predicted_ = _label_decode_covariates(
            self.predicted_, unique(self.covariates + self.eliminate_variance), self.encoders,
        )
        if self.reindexed:
            self.predicted_ = self._clean_index(self.predicted_)
        return self.predicted_

    def transform(self, df):
        """Predict regression target for df.

        The predicted regression target of an input sample is computed as the
        mean predicted regression targets of the trees in the forest.

        Parameters
        ----------
        df : NDFrame of shape [n_samples, n_features]
            Pandas dataframe with features, regression_features and covariates.

        Returns
        -------
        y : NDFrame of shape [n_samples, n_features]
            Data harmonized with Neuroharmony.
        """
        return self.predict(df.copy())

    def refit(self, df):
        """Fit a trained model with a new dataset.

        Parameters
        ----------
        df : NDFrame of shape [n_samples, n_features]
            Pandas dataframe with features, regression_features and covariates.
        """
        # Check data
        self._check_trained_model()
        self._check_data(df.copy())
        df = self._check_index(df.copy())
        self._check_training_ranges(df.copy())
        df, self.encoders = _label_encode_covariates(df.copy(), unique(self.covariates + self.eliminate_variance))
        self.models_by_feature_[self.features[0]]._check_is_fitted("predict")

        X_train_split, y_train_split = self._run_combat(df.copy())
        desc = "Retraining Neuroharmony hyperparameters: "
        for var in tqdm(self.features, desc=desc):
            self.models_by_feature_[var].best_estimator_[2].warm_start = False
            self.models_by_feature_[var].best_estimator_.fit(
                X_train_split[self.regression_features + [var]], y_train_split[var],
            )
        return self

    def _check_trained_model(self):
        if hasattr(self, "models_by_feature_"):
            return self
        else:
            raise NotFittedError(
                "This Neuroharmony instance is not fitted yet. Call 'fit' with appropriate arguments"
                " before using this estimator."
            )

    def _check_vars(self, df, vars):
        vars = Series(vars)
        # verify all needed variables are present
        is_feature_present = vars.isin(df.columns)
        missing_features_str = "Missing features: %s" % ", ".join(vars[~is_feature_present])
        self.numeric_features = df.select_dtypes(include=number).columns.tolist()
        assert is_feature_present.all(), ValueError(missing_features_str)

    def _check_training_ranges(self, df):
        vars = self.features + self.regression_features
        self.numeric_features = df[vars].select_dtypes(include=number).columns.tolist()
        numeric_vars = [var for var in vars if var in self.numeric_features]
        self.coverage_ = concat(
            [df[numeric_vars].min(skipna=True), df[numeric_vars].max(skipna=True)], axis=1, keys=["min", "max"],
        )

    def _check_prediction_ranges(self, df):
        vars = self.features + self.regression_features
        numeric_vars = [var for var in vars if var in self.numeric_features]
        self.prediction_is_covered_ = (
            df[numeric_vars]
            .apply(
                lambda column: column.between(self.coverage_["min"][column.name], self.coverage_["max"][column.name],)
            )
            .all(axis=1)
        )
        if not self.prediction_is_covered_.all():
            warn(
                "Some of the subject are out of the training range."
                "See Neuroharmony.subjects_out_of_range_ for a list of the affected subjects."
            )
            self.subjects_out_of_range_ = self.prediction_is_covered_[~self.prediction_is_covered_].index.tolist()

    def _check_data(self, df):
        type_error = "Input data should be a pandas dataframe (NDFrame)."
        assert isinstance(df.copy(), NDFrame), TypeError(type_error)
        self._check_vars(df.copy(), self.features)
        self._check_vars(df.copy(), self.covariates)
        self._check_vars(df.copy(), self.eliminate_variance)

    def _check_index(self, df):
        if df.index.duplicated().sum():
            df.index = [f"{index}_TMP_{i}" for i, index in enumerate(df.index)]
            self.reindexed = True
        return df

    def _clean_index(self, df):
        df.index = [index.split("_TMP_")[0] for index in df.index]
        return df

    def _get_pca_n_componets(self, X):
        X = self.scaler.fit_transform(X)
        self.decomposition.set_params(n_components=X.shape[1])
        self.decomposition.fit(X)
        n = next(i for i, x in enumerate(self.decomposition.explained_variance_ratio_) if x < 0.01)
        return [n + 1]

    def _clean_bad_pca_parameters(self, X):
        n_vars = len(self.regression_features) + 1
        if "PCA__n_components" in self.param_distributions.keys():
            if any([n_components > n_vars for n_components in self.param_distributions["PCA__n_components"]]):
                self.param_distributions["PCA__n_components"] = [
                    n_components
                    for n_components in self.param_distributions["PCA__n_components"]
                    if n_components > n_vars
                ]
                warn("Decomposition n_components > n_features are excluded from the parameters search.")
        else:
            self.param_distributions["PCA__n_components"] = self._get_pca_n_componets(X)

    def _random_search_with_leave_one_group_out_cv(self, X, y, groups):
        if self.decomposition.__class__.__name__ == "PCA":
            self._clean_bad_pca_parameters(X)
        self.leaveonegroupout_ = LeaveOneGroupOut()
        self.cv = list(self.leaveonegroupout_.split(X, y, groups))
        self.pipeline = Pipeline(
            steps=[
                (self.scaler.__class__.__name__, self.scaler),
                (self.decomposition.__class__.__name__, self.decomposition),
                (self.estimator.__class__.__name__, self.estimator),
            ]
        )
        self.pipeline.set_params(**self.pipeline_args)
        self.randomized_search_cv = RandomizedSearchCV(
            self.pipeline, param_distributions=self.param_distributions, cv=self.cv
        )
        self.randomized_search_cv.set_params(**self.randomized_search_args)
        self.randomized_search_cv.fit(X, y)
        return self.randomized_search_cv

    def _reconstruct_original_fieds(self, df, harmonized, extra_vars):
        """Concatenate ComBat data with the original data fields."""
        if df.index.name is not None:
            index_name = df.index.name
        else:
            index_name = "subject_index"
        if df.index.name in df.columns:
            index_name = df.index.name + "_y"
        harmonized = DataFrame(harmonized, index=df.index, columns=self.features)
        harmonized.index.rename(index_name, inplace=True)
        df = df.loc[harmonized.index][extra_vars].reset_index().copy()
        harmonized = harmonized.reset_index()
        return merge(harmonized, df, how="inner", on=index_name).set_index(index_name)

    def _train_neurofind(self, estimator=RandomForestRegressor()):
        return estimator.__class__.__name__

    def _run_combat(self, df):
        self.extra_vars = df.columns[~df.columns.isin(self.features)]
        combat = ComBat(self.features, self.covariates, self.eliminate_variance)
        self.X_harmonized_ = combat.transform(df.copy())
        self.X_harmonized_ = _label_decode_covariates(
            self.X_harmonized_, unique(self.covariates + self.eliminate_variance), self.encoders,
        )
        self.X_harmonized_.drop_duplicates(inplace=True)
        delta = df[self.features].subtract(self.X_harmonized_[self.features])
        # if delta.shape != df.shape:
        #     raise ValueError(f'DELTA = {delta.shape}, DF = {df.shape}, xh = {self.X_harmonized_.shape}')
        # y_train_split = merge(delta, df[self.extra_vars], how="inner", on=index_name).set_index(index_name)
        y_train_split = concat([delta, df[self.extra_vars]], axis=1, sort=False).dropna()
        X_train_split = df.loc[y_train_split.index]
        return X_train_split, y_train_split



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Setup python package."""

from setuptools import setup, find_packages


with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("requirements.txt") as requirements:
    requirements = requirements.readlines()

EXTRAS_REQUIRE = {
    "tests": ["pytest", "pytest-cov"],
    "docs": ["sphinx", "sphinx-gallery", "sphinx_rtd_theme", "numpydoc", "matplotlib"],
}

setup(
    author_email="rafaelagd@gmail.com",
    author="Rafael Garcia-Dias",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="A tool to perform Freesurfer volume Harminization in unseen scanner.",
    entry_points={"console_scripts": ["mriqc-run=neuroharmony.models.mriqc:main"]},
    extras_require=EXTRAS_REQUIRE,
    include_package_data=True,
    install_requires=requirements,
    keywords="Harminization, MRI, data science",
    license="MIT license",
    long_description=readme,
    name="Neuroharmony",
    package_dir={
        "neuroharmony": "neuroharmony",
        "models": "neuroharmony.models",
        "neuroCombat": "neuroharmony.models.neuroCombat",
    },
    packages=find_packages(),
    python_requires=">=3.6",
    test_suite="pytest",
    url="https://github.com/garciadias/Neuroharmony",
    version="0.1.10",
    zip_safe=False,
)



"""Combining tools."""
from pathlib import Path
import warnings

import pandas as pd

from neuroharmony.data.collect_tools import find_all_files_by_name
from neuroharmony.data.rois import rois


def _files_exists(directory_path, file_pattern):
    """Verify if a file exists and is the unique file of that kind in the folder."""
    file_search = find_all_files_by_name(directory_path, file_pattern)
    if len(file_search) == 0:
        return False
    elif len(file_search) == 1:
        return file_search[0]
    else:
        return file_search
        warnings.warn("There are more than one %s file in this site." % file_pattern)


def combine_freesurfer(freesurfer_path):
    """Combine aparc and aseg data files in a single csv files.

    It uses the list in columns_name.list file to select the relevant features in the freesurfer output.

    Parameters
    ----------
    freesurfer_path : string
        The path to the subject directories.

    Returns
    -------
    combined : NDFrame of shape [n_subjects, n_features]
        A DataFrame with the FreeSurfer volumetric data for each subject.
    """
    aseg_stats = pd.read_csv(f"{freesurfer_path}/aseg_stats.txt", delimiter="\t")
    lh_aparc_stats = pd.read_csv(f"{freesurfer_path}/lh_aparc_stats.txt", delimiter="\t")
    rh_aparc_stats = pd.read_csv(f"{freesurfer_path}/rh_aparc_stats.txt", delimiter="\t")

    combined = pd.merge(aseg_stats, lh_aparc_stats, left_on="Measure:volume", right_on="lh.aparc.volume")
    combined = pd.merge(combined, rh_aparc_stats, left_on="Measure:volume", right_on="rh.aparc.volume")
    combined.rename(columns={"Measure:volume": "participant_id"}, inplace=True)
    combined.participant_id = combined.participant_id.str.replace("/", "")
    combined = combined.set_index("participant_id")[rois]
    return combined


def combine_mriqc(mri_path=None, group_path=None, mclf_path=None):
    """Combine group_T1w and mclf files from the MRIQC output.

    It uses the list in columns_name.list file to select the relevant features in the freesurfer output.

    Parameters
    ----------
    mri_path: string
        The path to a BIDS directory.

    Returns
    -------
    combined : NDFrame of shape [n_subjects, n_features]
        A DataFrame with the MRIQC information for each subject.
    """
    if group_path is None:
        iqm_path = _files_exists(mri_path, "group_T1w.tsv")
    if mclf_path is None:
        pred_path = _files_exists(mri_path, "mclf*")
    if not iqm_path or not pred_path:
        raise FileNotFoundError(f"MRIQC files not found in {mri_path}")
    if isinstance(pred_path, list):
        pred_path = "\n".join(pred_path)
        raise TypeError(f"There are multiple mclf*.csv files at {mri_path}.\n\
                          Rename undesired mclf files you want to ignore or specify mclf_path.\n\
                          MCLF files found:\n{pred_path}")
    iqm = pd.read_csv(iqm_path, header=0, sep="\t")
    pred = pd.read_csv(pred_path, header=0)
    pred["bids_name"] = "sub-" + pred.subject_id + "_T1w"
    pred.drop(columns="subject_id", inplace=True)
    combined = pd.merge(iqm, pred, left_on="bids_name", right_on="bids_name")
    combined.rename(columns={"bids_name": "participant_id"}, inplace=True)
    combined.participant_id = [subject_line[0] for subject_line in combined.participant_id.str.split("_")]
    combined = combined.set_index("participant_id")
    return combined


class Site(object):
    """A class for site definition.

    Tools for collecting and combining data from a site given a path to the site data.

    Parameters
    ----------
    dir_path: pathlib.PosixPath
        Path to the data of the site. The following files are required: freesurferData.csv, participants.tsv,
        group_T1w.tsv, mclf*.csv, and Qoala*.csv. These files should be inside scanner folders.

    Attributes
    ----------
    data: NDFrame
        Dataframe containing the combined data of all scanners in the site.

    Examples
    --------
    >>> ixi = Site(Path(data/raw/IXI))
    >>> ixi.data

    +---------------+-------------+--------------+-------------+----------+---+----+
    |               |3rd-Ventricle|4th-Ventricle | Age         |Brain-Stem|...|site|
    +---------------+-------------+--------------+-------------+----------+---+----+
    |participant_id |                                                              |
    +---------------+-------------+--------------+-------------+----------+---+----+
    |sub-035-0      |0.000528     | 0.000607     | 37.144422   |0.015725  |...|IXI |
    +---------------+-------------+--------------+-------------+----------+---+----+
    |sub-230-0      |0.000751     | 0.001479     | 21.152635   |0.013917  |...|IXI |
    +---------------+-------------+--------------+-------------+----------+---+----+
    |sub-231-0      |0.000861     | 0.001068     | 58.992471   |0.012547  |...|IXI |
    +---------------+-------------+--------------+-------------+----------+---+----+
    |sub-232-0      |0.000478     | 0.000502     | 28.810404   |0.013322  |...|IXI |
    +---------------+-------------+--------------+-------------+----------+---+----+
    |sub-233-0      |0.000420     | 0.000725     | 26.754278   |0.014778  |...|IXI |
    +---------------+-------------+--------------+-------------+----------+---+----+
    >>> ixi.SCANNER01.data

    +---------------+----------------------+-----------------+---+-------------+
    |               |Left-Lateral-Ventricle|Left-Inf-Lat-Vent|...|scanner      |
    +---------------+----------------------+-----------------+---+-------------+
    |participant_id |                                                          |
    +---------------+----------------------+-----------------+---+-------------+
    |sub-002-2	    |      0.003274	       |     0.000123	 |...|IXI-SCANNER01|
    +---------------+----------------------+-----------------+---+-------------+
    |sub-016-2	    |      0.015889	       |     0.000380	 |...|IXI-SCANNER01|
    +---------------+----------------------+-----------------+---+-------------+
    |sub-017-2	    |      0.007326	       |     0.000148	 |...|IXI-SCANNER01|
    +---------------+----------------------+-----------------+---+-------------+
    |...	        |         ...	       |       ...       |...|     ...	   |
    +---------------+----------------------+-----------------+---+-------------+
    |sub-582-2	    |      0.002624	       |     0.000146	 |...|IXI-SCANNER01|
    +---------------+----------------------+-----------------+---+-------------+
    |sub-584-2	    |      0.004854	       |     0.000188	 |...|IXI-SCANNER01|
    +---------------+----------------------+-----------------+---+-------------+
    |sub-585-2	    |      0.004373	       |     0.000218	 |...|IXI-SCANNER01|
    +---------------+----------------------+-----------------+---+-------------+
    """

    def __init__(self, dir_path, merge_on="image_id", renamed_id="participant_id"):
        self.dir_path = dir_path
        self.merge_on = merge_on
        self.renamed_id = renamed_id
        self.name = dir_path.name
        self._get_scanners()

    def _get_scanners(self):
        """
        Collect data for the site.

        This method verifies if the site has multiples scanners and combine the data from freesurfer, MRIQC and Qoala.
        If the data is not available it returns an empty dataframe.

        """
        self.dir_name = self.dir_path.name
        subdirs = [subpath.name for subpath in self.dir_path.glob("*") if subpath.is_dir()]
        if len(subdirs) == 0:
            self.scanner_list = []
            self._get_files()
            if self._is_complete():
                self._load_files()
                self._combine_files()
                self.data["scanner"] = self.name
                self.data.index = self.data[self.renamed_id] + "-00"
                self.complete_scanners = [self.name]
            else:
                self.data = pd.DataFrame()
        else:
            self.scanner_list = [subdir.replace(f"{self.dir_name}-", "") for subdir in subdirs]
            for scanner_name in subdirs:
                scanner = Scanner(
                    self.dir_path / Path(scanner_name), merge_on=self.merge_on, renamed_id=self.renamed_id
                )
                scanner_name = scanner_name.replace(f"{self.dir_name}-", "")
                setattr(self, scanner_name, scanner)
                scanner._get_files()
            self._combine_all_scanners()
        self.data["site"] = self.name

    def _files_exists(self, directory_path, file_pattern):
        """Verify if a file exists and is the unique file of that kind in the folder."""
        file_search = find_all_files_by_name(directory_path, file_pattern)
        if len(file_search) == 0:
            return False
        elif len(file_search) == 1:
            return file_search[0]
        elif len(file_search) >= 1:
            error_str = "\n".join([str(p) for p in file_search])
            warnings.warn(f"There are multiple mclf*.csv files at {directory_path}.\n"
                          "Rename undesired mclf files you want to ignore or specify mclf_path.\n"
                          f"MCLF files found:\n{error_str}\nUsing: {file_search[0]}")
            return file_search[0]

    def _get_files(self):
        """Verify if each of the files exist and is unique."""
        self.freesurferData_path = self._files_exists(self.dir_path, "freesurferData.csv")
        self.participants_path = self._files_exists(self.dir_path, "participants.tsv")
        self.iqm_path = self._files_exists(self.dir_path, "group_T1w.tsv")
        self.pred_path = self._files_exists(self.dir_path, "mclf*csv")
        self.qoala_path = self._files_exists(self.dir_path, "Qoala*.csv")

    def _is_complete(self):
        """Verify if all necessary files are there."""
        return all([self.freesurferData_path, self.participants_path, self.iqm_path, self.pred_path, self.qoala_path])

    def _load_files(self):
        self.freesurferData = pd.read_csv(self.freesurferData_path, header=0)
        self.participants = pd.read_csv(self.participants_path, header=0, sep="\t")
        self.iqm = pd.read_csv(self.iqm_path, header=0, sep="\t")
        self.pred = pd.read_csv(self.pred_path, header=0)
        self.qoala = pd.read_csv(self.qoala_path, header=0)

    def _combine_files(self):
        if all(
            self.merge_on in dataframe
            for dataframe in [self.freesurferData, self.participants, self.iqm, self.pred, self.qoala]
        ):
            self.data = pd.merge(self.participants, self.freesurferData, on=self.merge_on, how="inner")
            self.data = pd.merge(self.data, self.iqm, on=self.merge_on, how="inner")
            self.data = pd.merge(self.data, self.pred, on=self.merge_on, how="inner")
            self.data = pd.merge(self.data, self.qoala, on=self.merge_on, how="inner")
            self.data[self.renamed_id] = self.data[self.merge_on]
        else:
            warnings.warn("The field %s is not found in %s." % (self.merge_on, self.dir_path.name))
            self.data = pd.DataFrame([""], index=[""], columns=[self.renamed_id])

    def _combine_all_scanners(self):
        """If there are multiple scanners on a site they will be combined in the attribute 'data'."""
        n_scanners = len(self.scanner_list)
        participant_id_format = "-%%0%dd" % len(str(n_scanners))
        for scanner_id, scanner_name in enumerate(self.scanner_list):
            if getattr(self, scanner_name)._is_complete():
                getattr(self, scanner_name)._load_files()
                getattr(self, scanner_name)._combine_files()
                scanner_data = getattr(self, scanner_name).data
                if len(scanner_data) > 0:
                    scanner_data["scanner"] = "%s-%s" % (self.name, scanner_name)
                    try:
                        id_appendix = participant_id_format % scanner_id
                        scanner_data.index = scanner_data[self.renamed_id] + str(id_appendix)
                    except TypeError:
                        print(self.dir_path.name)
        self.complete_scanners = [
            scanner_name for scanner_name in self.scanner_list if getattr(self, scanner_name)._is_complete()
        ]
        if len(self.complete_scanners) > 0:
            self.data = pd.concat([getattr(self, scanner).data for scanner in self.complete_scanners], sort=True)
        else:
            self.data = pd.DataFrame()


class Scanner(Site):
    """
    A class for scanner definition.

    Tools for collecting and combining data from a single scanner given a path to the scanner data.

    Parameters
    ----------
    dir_path: pathlib.PosixPath
        Path to the data of the scanner. The following files are required: freesurferData.csv, participants.tsv,
        group_T1w.tsv, mclf*.csv, and Qoala*.csv.

    Attributes
    ----------
    data: NDFrame
        Dataframe containing the combined data of all scanners in the site.

    Examples
    --------
    >>> scanner_01 = Site(Path(data/raw/IXI/SCANNER01))
    >>> scanner_01.data
                Left-Lateral-Ventricle	 Left-Inf-Lat-Vent ... scanner
    participant_id
    sub-002-2	0.003274	             0.000123	... IXI-SCANNER01
    sub-016-2	0.015889	             0.000380	... IXI-SCANNER01
    sub-017-2	0.007326	             0.000148	... IXI-SCANNER01
    ...	        ...	                     ...	    ...	...
    sub-582-2	0.002624	             0.000146	... IXI-SCANNER01
    sub-584-2	0.004854	             0.000188	... IXI-SCANNER01
    sub-585-2	0.004373	             0.000218	... IXI-SCANNER01

    [313 rows × 179 columns]

    """

    def __init__(self, dir_path, merge_on="image_id", renamed_id="participant_id"):
        self.dir_path = dir_path
        self.name = dir_path.name
        self.merge_on = merge_on
        self.renamed_id = renamed_id


class DataSet(Site):
    """
    A class for dataset definition.

    Tools for collecting and combining data from all sites given a path to a BIDS dataset.

    Parameters
    ----------
    dir_path: pathlib.PosixPath
        Path to the data of the site. The following files are required: freesurferData.csv, participants.tsv,
        group_T1w.tsv, mclf*.csv, and Qoala*.csv. These files should be inside scanner folders which should be inside
        site folders.

    Attributes
    ----------
    data: NDFrame
        Dataframe containing the combined data of all scanners in the site.

    Examples
    --------
    >>> DATASET = DataSet(Path(data/raw/))
    >>> Data.data
                         3rd-Ventricle  4th-Ventricle  Age          Brain-Stem ... site
    participant_id
    sub-035-00-00        0.000528       0.000607       37.144422    0.015725   ... IXI
    sub-230-00-00        0.000751       0.001479       21.152635    0.013917   ... IXI
    sub-231-00-00        0.000861       0.001068       58.992471    0.012547   ... IXI
    ...                  ...            ...            ...          ...        ... ...
    sub-45-00-33	     0.000570	    0.001581	   21.0	        0.011838   ... MoralJudgment
    sub-46-00-33	     0.000552	    0.001528	   36.0	        0.013947   ... MoralJudgment
    sub-47-00-33	     0.000609	    0.000893	   20.0	        0.013515   ... MoralJudgment

    [9303 rows × 180 columns]

    >>> DATASET.IXI.data
                         3rd-Ventricle  4th-Ventricle  Age          Brain-Stem ... site
    participant_id
    sub-035-0            0.000528       0.000607       37.144422    0.015725   ... IXI
    sub-230-0            0.000751       0.001479       21.152635    0.013917   ... IXI
    sub-231-0            0.000861       0.001068       58.992471    0.012547   ... IXI
    sub-232-0            0.000478       0.000502       28.810404    0.013322   ... IXI
    sub-233-0            0.000420       0.000725       26.754278    0.014778   ... IXI

    [535 rows x 180 columns]

    >>> DATASET.IXI.SCANNER01.data
                Left-Lateral-Ventricle	 Left-Inf-Lat-Vent ... scanner
    participant_id
    sub-002-2	0.003274	             0.000123	... IXI-SCANNER01
    sub-016-2	0.015889	             0.000380	... IXI-SCANNER01
    sub-017-2	0.007326	             0.000148	... IXI-SCANNER01
    ...	        ...	                     ...	    ...	...
    sub-582-2	0.002624	             0.000146	... IXI-SCANNER01
    sub-584-2	0.004854	             0.000188	... IXI-SCANNER01
    sub-585-2	0.004373	             0.000218	... IXI-SCANNER01

    [313 rows × 179 columns]

    """

    def __init__(self, dir_path, merge_on="image_id", renamed_id="participant_id"):
        self.dir_path = dir_path
        self.name = "All sites"
        self.merge_on = merge_on
        self.renamed_id = renamed_id
        self._get_sites()
        self._combine_all_sites()

    def _get_sites(self):
        self.site_paths = [subpath for subpath in self.dir_path.glob("*") if subpath.is_dir()]
        self.sites = [site_path.name for site_path in self.site_paths]
        self.not_empty_sites = []
        for site_path in self.site_paths:
            site = Site(site_path, merge_on=self.merge_on, renamed_id="participant_id")
            setattr(self, site.name, site)
            self.not_empty_sites.append(site.name)

    def _combine_all_sites(self):
        n_sites = len(self.not_empty_sites)
        participant_id_format = "-%%0%dd" % len(str(n_sites))
        self.data = []
        if n_sites > 0:
            for site_id, site_name in enumerate(self.not_empty_sites):
                id_appendix = participant_id_format % site_id
                site = getattr(self, site_name)
                site.data.index = site.data.index + id_appendix
                self.data.append(site.data)
            self.data = pd.concat(self.data, sort=True)
        else:
            self.data = False


if __name__ == "__main__":
    DATASET = DataSet(Path("data/raw/"))
    DATASET.data.to_csv("data/processed/dataset.csv")



"""Tests for harmonization with Combat."""
from collections import namedtuple

from pandas.core.generic import NDFrame
from pandas import Series
from pathlib import Path
import pytest
from scipy.special import comb

from neuroharmony.data.combine_tools import DataSet
from neuroharmony.models.harmonization import ComBat, exclude_single_subject_groups
from neuroharmony.models.metrics import ks_test_grid
from neuroharmony.data.rois import rois
from neuroharmony.models.neuroCombat import neuroCombat


def compare_dfs(ks_original, ks_harmonized):
    """Compare two dictionaries of dataframes to measure the success of harmonization."""
    vars = ks_original.keys()
    var_improved = Series(index=vars, dtype="bool")
    for var in vars:
        improved = ks_original[var] < ks_harmonized[var]
        var_improved.loc[var] = improved.sum().sum()
    return var_improved


@pytest.fixture(scope="session")
def resources(tmpdir_factory):
    """Set up."""
    r = namedtuple("resources", "data_path")
    r.data_path = "data/raw/IXI"
    r.data = DataSet(Path(r.data_path)).data
    r.features = rois[:3]
    r.covariates = ["Gender", "scanner", "Age"]
    r.data = r.data[~r.data[r.covariates].isna().any(axis=1)]
    r.data.Age = r.data.Age.astype(int)
    r.data = r.data.drop_duplicates()
    r.data = r.data[~r.data.isna().any(axis=1)]
    r.data = exclude_single_subject_groups(r.data, r.covariates)
    r.eliminate_variance = ["scanner"]
    r.n_scanners = len(r.data.scanner.unique())
    return r


def test_combat_is_functional(resources):
    """Test ComBat harmonization returns a NDFrame with no NaN values and conserves scanner column format."""
    combat = ComBat(resources.features, resources.covariates, resources.eliminate_variance)
    data_harmonized = combat.transform(resources.data[resources.features + resources.covariates].copy())
    assert isinstance(data_harmonized, NDFrame)
    assert not data_harmonized.isna().any(axis=1).any()
    assert_message = "Scanner field is not string data_harmonized.scanner[0] = %s" % data_harmonized.scanner[0]
    assert isinstance(data_harmonized.scanner[0], str), assert_message


def test_harmonization_works(resources):
    """Test harmonization with ComBat using the Kolmogorov-Smirnov test."""
    combat = ComBat(resources.features, resources.covariates, resources.eliminate_variance)
    data_harmonized = combat.transform(resources.data[resources.features + resources.covariates])
    KS_harmonized = ks_test_grid(data_harmonized, resources.features, "scanner")
    assert isinstance(KS_harmonized, dict)
    assert isinstance(KS_harmonized[resources.features[0]], NDFrame)
    assert KS_harmonized[resources.features[0]].shape == (resources.n_scanners, resources.n_scanners)
    KS_original = ks_test_grid(resources.data, resources.features, "scanner")
    assert (compare_dfs(KS_original, KS_harmonized) >= comb(resources.n_scanners, 2) - 1).all()



from neuroharmony.data.collect_tools import fetch_mri_data, fetch_sample, fetch_trained_model
from neuroharmony.data.combine_tools import combine_freesurfer, combine_mriqc
from neuroharmony.models.harmonization import ComBat, Neuroharmony, exclude_single_subject_groups
from neuroharmony.models.mriqc import run_mriqc
from neuroharmony.models.metrics import ks_test_grid
